/*
    Copyright (C) 2010  Thomas Schrøder

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    LICENSE.txt file for license details.
*/

#include "rumd_data.h"
#include <unistd.h>

void CalcR2(Conf *C0, Conf *Ct, double *R2,  double *R4, double *Fs, double *Fs2, unsigned int num_types, double *q) {
  unsigned int num_part = C0->num_part;
  float L = C0->metaData.L[0];

  double *Fs_;
  Fs_ = new double[num_types];

  for (unsigned int i=0; i<num_types; i++) Fs_[i] = 0.0;

  for (unsigned int i=0; i<num_part; i++) {
    unsigned int Type = C0->P[i].MyType;
    assert(Type<num_types);
    assert(Type==Ct->P[i].MyType);
    double Rsq = C0->P[i].Rsq(&Ct->P[i], L, L, L );
    R2[Type]  += Rsq;
    R4[Type]  += Rsq*Rsq;
    Fs_[Type] += C0->P[i].Fsq(&Ct->P[i], L, L, L, q[Type]);
  }
  for (unsigned int i=0; i<num_types; i++) {
    Fs[i] += Fs_[i];
    Fs2[i] += Fs_[i]*Fs_[i];
  }
  delete[] Fs_;
}


int main(int argc, char *argv[])
{
  unsigned int first_block=0;
  unsigned int last_block;

  static const char *usage = "Usage: %s [-h]\n";
  int opt;
  while ((opt = getopt(argc, argv, "h")) != -1) {
    switch (opt) {
    case 'h':
      fprintf(stdout, usage, argv[0]);
      exit(EXIT_SUCCESS);
    case '?':
      fprintf(stderr, "%s: unknown option: -%c\n", argv[0], optopt);
      fprintf(stderr, usage, argv[0]);
      exit(EXIT_FAILURE);
      break;
    case ':':
      fprintf(stderr, "%s: missing option argument for option: -%c\n", argv[0], optopt);
      fprintf(stderr, usage, argv[0]);
      exit(EXIT_FAILURE);
      break;
    default:
      fprintf(stderr, usage, argv[0]);
      exit(EXIT_FAILURE);
    }
  }

  // Change to more user friendly option reader:
  //sscanf(argv[1], "%d", &first_block);
  //sscanf(argv[2], "%d", &last_block);

  Conf *C0 = new Conf();
  Conf *Ct = new Conf();

  gzFile gzfp;
  FILE *fp;

  fp=fopen("TrajectoryFiles/LastBlockSaved","r");
  if (!fp){
    fprintf (stderr, "Error opening file TrajectoryFiles/LastBlockSaved: %s\n", strerror(errno));
    exit(EXIT_FAILURE);
  }
  fscanf(fp, "%u", &last_block);
  fclose(fp);

  // Read metaData from first configuration in first block
  char *C0Name;
  asprintf(&C0Name,"./TrajectoryFiles/block%.4u.xyz.gz", first_block);
  printf("Reading block %u: %s\n\n", first_block, C0Name);
  gzfp=gzopen(C0Name,"r");
  if (!gzfp){
    fprintf (stderr, "rumd_msd: Error opening file %s: %s\n",C0Name, strerror(errno));
    exit(EXIT_FAILURE);
  }
  C0->Read(gzfp,C0Name,1);
  gzclose(gzfp);
  free(C0Name);

  // Store the metaData we need
  // float L  = C0->metaData.L[0];     // Assuming Lx=Ly=Lz for now...
  float dt = C0->metaData.dt;
  unsigned int num_part = C0->num_part;
  unsigned int num_types = C0->metaData.numTypes;
  unsigned int MaxIndex = C0->metaData.logLin.maxIndex;

  double *qvalues;
  int *NoPartOfType;
  qvalues = new double[num_types];
  NoPartOfType = new int[num_types];

  fp=fopen("qvalues.dat","r");
  if (!fp){
    fprintf (stderr, "Error opening file qvalues.dat: %s\n", strerror(errno));
    exit(EXIT_FAILURE);
  }
  printf("\nReading q-values: ");
  for (unsigned int i=0; i<num_types; i++) {
    fscanf(fp, "%lf", &qvalues[i]);
    printf("%f ", qvalues[i]);
    assert(qvalues[i]>0);
    NoPartOfType[i] = 0;
  }
  printf("\n\n");
  fclose(fp);

  // Read and print the number of particles of each type
  for (unsigned int i=0; i<num_part; i++) {
    unsigned int Type = C0->P[i].MyType;
    assert(Type<num_types);
    NoPartOfType[Type]++;
  }
  for (unsigned int i=0; i<num_types; i++) {
    printf("Particles of type %d: %d\n", i, NoPartOfType[i]);
  }

  // Setup arrays to collect statistics
  unsigned long int *Time;
  unsigned long int *Count;
  double **R2;
  double **R4;
  double **Fs;
  double **Fs2;

  int MaxDataIndex = MaxIndex + int(log2(last_block-first_block+1)+1);
  //printf("\nData structure info: %d %d %d %d\n", num_part, num_types, MaxIndex, MaxDataIndex);

  Time = new unsigned long int[MaxDataIndex];
  R2  = new double*[MaxDataIndex];
  R4  = new double*[MaxDataIndex];
  Fs  = new double*[MaxDataIndex];
  Fs2 = new double*[MaxDataIndex];
  Count = new unsigned long int[MaxDataIndex];
  for (int i=0; i<MaxDataIndex; i++ ) {
    R2[i]  = new double[num_types];
    R4[i]  = new double[num_types];
    Fs[i]  = new double[num_types];
    Fs2[i] = new double[num_types];
    Count[i] = 0;
    for (unsigned int j=0; j<num_types; j++) {
      R2[i][j] = 0.0;
      R4[i][j] = 0.0;
      Fs[i][j] = 0.0;
      Fs2[i][j] = 0.0;
    }
  }

  printf("\nReading block: ");
  for (unsigned int i=first_block; i<=last_block; i++) {

    // Read first configuration of block 'i'
    char *C0Name;
    asprintf(&C0Name,"./TrajectoryFiles/block%.4u.xyz.gz",i);
    if ((i-first_block)%10==0) printf("%u ", i);
    fflush(stdout);
    gzfp=gzopen(C0Name,"r");
    C0->Read(gzfp,C0Name,0);

    // Check that relevant metaData didn't change
    assert(C0->metaData.dt       == dt);
    assert(C0->num_part          == num_part);
    assert(C0->metaData.numTypes == num_types);
    assert(MaxIndex == C0->metaData.logLin.maxIndex);

    // Analyze configurations in the same block-file
    for (unsigned int j=1; j<=MaxIndex; j++) {
      Ct->Read(gzfp,C0Name,0);
      Count[j-1]++;
      CalcR2(C0, Ct, R2[j-1], R4[j-1], Fs[j-1], Fs2[j-1], num_types, qvalues);
    }
    gzclose(gzfp);
    free(C0Name);

    // Analyze configurations in the later block-files
    for (unsigned int j=1; i+(1<<j)<=last_block; j++) {
      char *C0Name;
      asprintf(&C0Name,"./TrajectoryFiles/block%.4u.xyz.gz",i+(1<<j));
      gzfp=gzopen(C0Name,"r");
      Ct->Read(gzfp,C0Name,0);
      int jj = MaxIndex+j-1;
      assert(jj<MaxDataIndex);
      Count[jj]++;
      CalcR2(C0, Ct, R2[jj], R4[jj], Fs[jj], Fs2[jj], num_types, qvalues);
      gzclose(gzfp);
      free(C0Name);
     }
  }
  printf("\n");

  fp = fopen("msd.dat","w");
  for (int i=0; i<MaxDataIndex; i++ ) {
    if (Count[i]>0) {
      fprintf(fp, "%f ", (1UL<<i)*dt);
      for (unsigned int j=0; j<num_types; j++) {
	fprintf(fp, "%f ", R2[i][j]/float(Count[i]*NoPartOfType[j]));
      }
      fprintf(fp, "\n ");
    }
  }
  fclose(fp);

  fp = fopen("alpha2.dat","w");
  for (int i=0; i<MaxDataIndex; i++ ) {
    if (Count[i]>0) {
      fprintf(fp, "%f ", (1UL<<i)*dt);
      for (unsigned int j=0; j<num_types; j++) {
	fprintf(fp, "%f ", 3.f/5.f*R4[i][j]/powf(R2[i][j],2.f)*float(Count[i]*NoPartOfType[j])-1.f);
      }
      fprintf(fp, "\n ");
    }
  }

  fp = fopen("Fs.dat","w");
  for (int i=0; i<MaxDataIndex; i++ ) {
    if (Count[i]>0) {
      fprintf(fp, "%f ", (1UL<<i)*dt);
      for (unsigned int j=0; j<num_types; j++) {
	fprintf(fp, "%f ", Fs[i][j]/float(Count[i]*NoPartOfType[j]));
      }
      fprintf(fp, "\n ");
    }
  }
  fclose(fp);

  fp = fopen("FsSq.dat","w");
  for (int i=0; i<MaxDataIndex; i++ ) {
    if (Count[i]>0) {
      fprintf(fp, "%f ", (1UL<<i)*dt);
      for (unsigned int j=0; j<num_types; j++) {
	fprintf(fp, "%f ", Fs2[i][j]/float(Count[i]*NoPartOfType[j]) );
      }
      fprintf(fp, "\n ");
    }
  }
  fclose(fp);

}
