/*
    Copyright (C) 2010  Thomas Schrøder

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    LICENSE.txt file for license details.
*/

#include "md_io_helpers.h"
#include <cstring>
#include <cassert>
#include <cerrno>
#include <cmath>
#include <zlib.h>

int main(int argc, char *argv[])
{
  unsigned int first_block=0;
  unsigned int last_block;
  bool verbose = false; //true;

  static const char *usage = "Usage: %s [-h]\n";
  int opt;
  while ((opt = getopt(argc, argv, "h")) != -1) {
    switch (opt) {
    case 'h':
      fprintf(stdout, usage, argv[0]);
      exit(EXIT_SUCCESS);
    case '?':
      fprintf(stderr, "%s: unknown option: -%c\n", argv[0], optopt);
      fprintf(stderr, usage, argv[0]);
      exit(EXIT_FAILURE);
      break;
    case ':':
      fprintf(stderr, "%s: missing option argument for option: -%c\n", argv[0], optopt);
      fprintf(stderr, usage, argv[0]);
      exit(EXIT_FAILURE);
      break;
    default:
      fprintf(stderr, usage, argv[0]);
      exit(EXIT_FAILURE);
    }
  }

  gzFile gzfp;
  FILE *fp;

  fp=fopen("TrajectoryFiles/LastBlockSaved","r");
  if (!fp){
    fprintf (stderr, "Error opening file TrajectoryFiles/LastBlockSaved: %s\n", strerror(errno));
    exit(1);
  }
  fscanf(fp, "%u", &last_block);
  fclose(fp);
  
  //read energy file for first block 
  
  char *E0_Name;
  asprintf(&E0_Name,"./TrajectoryFiles/energies%.4u.dat.gz", first_block);
  
  if(verbose) printf("Reading block %u: %s\n\n", first_block, E0_Name);
  gzfp=gzopen(E0_Name,"r");
  if (!gzfp){
    fprintf (stderr, "rumd_stats: Error opening file %s: %s\n",E0_Name, strerror(errno));
    exit(1);
  }
  
  char strWorkspace[256];
  unsigned int maxLen = sizeof(strWorkspace)/sizeof(char);
  // read the first line
  bool ok = gzgets(gzfp, strWorkspace, maxLen);
  if(!ok) {
    fprintf(stderr, "rumd_stats: Error reading first line of file %s\n", E0_Name);
    exit(1);
  }
  else {
    // replace final newline with a zero
    strWorkspace[strlen(strWorkspace)-1] = 0;
  }
  OutputFormat OF;
  EnergiesMetaReader(strWorkspace, &OF, verbose);
  
  gzclose(gzfp);
  free(E0_Name);

  int num_cols = 0;
  // allocate memory for statistics
  std::map<std::string, bool>::const_iterator it;
  for(it = OF.Start(); it != OF.End(); it++ )
    if(OF.Get(it->first)) num_cols += 1;
  if(verbose) printf("Number of columns is %d\n",num_cols);
  double* mean_vals = new double[num_cols];
  double* meanSq_vals = new double[num_cols];
  unsigned int count = 0;

  // now read all blocks
  for (unsigned int i=first_block; i<=last_block; i++) {
    char* E_Name;
    asprintf(&E_Name,"./TrajectoryFiles/energies%.4u.dat.gz", i);
    if(verbose) printf("Reading block %u: %s\n", i, E_Name);
    gzfp=gzopen(E_Name,"r");
    if (!gzfp){
      fprintf (stderr, "rumd_stats: Error opening file %s: %s\n",E_Name, strerror(errno));
      exit(1);
    }
    // read first line, but don't do anything with it (could perhaps
    // check for consistency ... )
    ok = gzgets(gzfp, strWorkspace, maxLen);
    if(!ok) {
      fprintf(stderr, "rumd_stats: Error reading first line of file %s\n", E_Name);
      exit(1);
    }    
    // read data lines
    ok = gzgets(gzfp, strWorkspace, maxLen);
    while(ok) {
      int offset = 0;
      int partial_offset = 0;
      double value = 0.;
      for(int idx = 0; idx < num_cols; ++idx) {
	sscanf(strWorkspace+offset, "%lf%n", &value, &partial_offset);
        offset += partial_offset;
	mean_vals[idx] += value;
	meanSq_vals[idx] += value*value;
      } 
      count += 1;
      ok = gzgets(gzfp, strWorkspace, maxLen);
    } // end while loop over lines in this file

    gzclose(gzfp);
    free(E_Name);    
  } // end loop over blocks

  printf("Total line count: %u\n", count);
  printf("quantity, mean value, variance, standard deviation\n");
  FILE* meanFile = fopen("energies_mean.dat","w");
  FILE* meanSqFile = fopen("energies_mean_sq.dat","w");
  FILE* varFile = fopen("energies_var.dat","w");
  
  fprintf(meanFile,"# ");
  fprintf(meanSqFile,"# ");
  fprintf(varFile,"# ");	
  for(it = OF.Start(); it != OF.End(); it++) 
    if(it->second){
      fprintf(meanFile," %s",OF.FileString(it->first).c_str());
      fprintf(varFile," %s",OF.FileString(it->first).c_str());   
    }
  fprintf(meanFile,"\n");
  fprintf(meanSqFile,"\n");
  fprintf(varFile,"\n");

  it = OF.Start(); 
  while(!it->second) it++;
  for(int idx = 0; idx < num_cols; ++idx) {
    mean_vals[idx] /= count;
    meanSq_vals[idx] /= count;
    double variance = (meanSq_vals[idx] - mean_vals[idx]*mean_vals[idx]);//*count/(count-1.);
    double std_dev = sqrt(variance);
    printf("%s\t%12.6g\t%12.6g\t%12.6g\n", OF.FileString(it->first).c_str(), mean_vals[idx], variance, std_dev );
    fprintf(meanFile," %12.10g",mean_vals[idx]);
    fprintf(meanSqFile," %12.10g",meanSq_vals[idx]);
    fprintf(varFile," %12.10g",variance);
    do { it++; } while (it != OF.End() && !it->second);
  }
  fprintf(meanFile,"\n");
  fprintf(varFile,"\n");
  fclose(meanFile);
  fclose(meanSqFile);
  fclose(varFile);
  assert(it == OF.End());

  // calculate statistics
  delete [] mean_vals;
  delete [] meanSq_vals;
}
