/*
    Copyright (C) 2010  Thomas Schrøder

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    LICENSE.txt file for license details.
*/
#include "md_IntegratorNVU.h"
#include "md_kernel.h"

//////////////////////////////////////////////////////////////////////////
///////////////////////// Constructors ///////////////////////////////////
//////////////////////////////////////////////////////////////////////////

IntegratorNVU::IntegratorNVU(Sample* newSample){
  S = newSample; // Set the sample.
  numberOfBlocks = 1;
  targetPotentialEnergy = 1.f;
  timeStep = 1.f;
  displacementLength = 0.1f;

  AllocateNVUState(numberOfBlocks, targetPotentialEnergy); 
}

IntegratorNVU::~IntegratorNVU(){
  FreeNVUState();
}

//////////////////////////////////////////////////////////////////////////
///////////////////////// Class Methods //////////////////////////////////
//////////////////////////////////////////////////////////////////////////

void IntegratorNVU::Integrate(){
}

// Constrains the motion on U.
void IntegratorNVU::CalculateConstraintForce(){
}

// Step length and potential energy correction. Reuses the summation arrays.
void IntegratorNVU::CalculateNumericalCorrection(){
}

void IntegratorNVU::AllocateNVUState(){
  numberOfBlocks = S->allocatedNumBlocks;
  FreeNVUState();
  AllocateNVUState(numberOfBlocks, targetPotentialEnergy);
}

// Private method.
void IntegratorNVU::AllocateNVUState( int nb, float targetPotentialEnergy ){
  // U(n-1).
  h_previousPotentialEnergy = (float*) malloc( sizeof(float) );
  h_previousPotentialEnergy[0] = S->num_part * targetPotentialEnergy; 
  cudaMalloc( (void**) &d_previousPotentialEnergy, sizeof(float) );  
  cudaMemcpy( d_previousPotentialEnergy, h_previousPotentialEnergy, sizeof(float), cudaMemcpyHostToDevice );
  
  // Constraint force.
  h_constraintForce = (float*) malloc( sizeof(float) );
  cudaMalloc( (void**) &d_constraintForce, sizeof(float) );  
  cudaMalloc( (void**) &d_totalConstraintForceDenominator, sizeof(float) );  
  
  // Numerical corrections CPU.
  h_constraintForceCorrection = (float*) malloc( sizeof(float) );
  h_constraintForceCorrection[0] = 0;
  h_displacementLengthCorrection = (float*) malloc( sizeof(float) );
  h_displacementLengthCorrection[0] = 1.f;
  
  // Numerical corrections GPU.
  cudaMalloc( (void**) &d_constraintForceCorrection, sizeof(float) );  
  cudaMalloc( (void**) &d_displacementLengthCorrection, sizeof(float) );  
  cudaMemcpy( d_displacementLengthCorrection, h_displacementLengthCorrection, sizeof(float), cudaMemcpyHostToDevice );

  // Allocate space on the GPU for the summation of the constraint force.
  cudaMalloc( (void**) &d_constraintForceNumerator, nb * sizeof(float) );  
  cudaMalloc( (void**) &d_constraintForceDenominator, nb * sizeof(float) );  

  // Allocate space for the summation of the system velocity on GPU.
  cudaMalloc( (void**) &d_particleVelocity, nb * sizeof(float4) );  

  // Euclidean configuration.
  h_euclideanConfiguration = (float4*) malloc( S->num_part * sizeof(float4) );
}

void IntegratorNVU::FreeNVUState(){
  free(h_previousPotentialEnergy);
  free(h_constraintForce);
  free(h_constraintForceCorrection);
  free(h_displacementLengthCorrection);
  free(h_euclideanConfiguration);
  cudaFree(d_previousPotentialEnergy);
  cudaFree(d_constraintForce);
  cudaFree(d_totalConstraintForceDenominator);
  cudaFree(d_constraintForceCorrection);
  cudaFree(d_displacementLengthCorrection);
  cudaFree(d_constraintForceNumerator);
  cudaFree(d_constraintForceDenominator);
  cudaFree(d_particleVelocity);
}

//////////////////////////////////////////////////////////////////////////
///////////////////////// Set Methods ////////////////////////////////////
//////////////////////////////////////////////////////////////////////////

void IntegratorNVU::SetTimeStep( float dt ){ timeStep = dt; }
void IntegratorNVU::SetDisplacementLength( float step ){ displacementLength = step; }
void IntegratorNVU::SetTargetPotentialEnergy( float U ){ targetPotentialEnergy = U; };
void IntegratorNVU::SetSystemVelocityToZero(){
}

void IntegratorNVU::SetEuclideanConfiguration(){
  S->particleData.CopyPosFromDevice();

  for(unsigned int i=0; i < S->num_part; i++){
    h_euclideanConfiguration[i] = S->particleData.h_r[i];
  }
}

void IntegratorNVU::SetEuclideanConfiguration(float4* c){
  for(unsigned int i=0; i < S->num_part; i++){
    h_euclideanConfiguration[i] = c[i];
  }
}

//////////////////////////////////////////////////////////////////////////
///////////////////////// Get Methods ////////////////////////////////////
//////////////////////////////////////////////////////////////////////////

float IntegratorNVU::GetTimeStep(){ return timeStep; }

float IntegratorNVU::GetSimulationTimeStepSq(){ 
  cudaMemcpy( h_constraintForce, d_constraintForce, sizeof(float), cudaMemcpyDeviceToHost );
  return h_constraintForce[0]; 
}

float IntegratorNVU::GetDisplacementLength(){ return displacementLength; }

float IntegratorNVU::GetSimulationDisplacementLength(){ 
  S->particleData.CopyVelFromDevice();
  
  float dispLength = 0;

  for(unsigned int i=0; i < S->num_part; i++){
    dispLength += S->particleData.h_v[i].x * S->particleData.h_v[i].x + S->particleData.h_v[i].y * S->particleData.h_v[i].y 
      + S->particleData.h_v[i].z * S->particleData.h_v[i].z;
  }

  return sqrtf(dispLength);
} 

float IntegratorNVU::GetConstraintForceCorrection(){ 
  cudaMemcpy( h_constraintForceCorrection, d_constraintForceCorrection, sizeof(float), cudaMemcpyDeviceToHost );
  return h_constraintForceCorrection[0]; 
};

float IntegratorNVU::GetDisplacementLengthCorrection(){ 
  cudaMemcpy( h_displacementLengthCorrection, d_displacementLengthCorrection, sizeof(float), cudaMemcpyDeviceToHost );
  return h_displacementLengthCorrection[0]; 
};

float IntegratorNVU::GetTargetPotentialEnergy(){ return targetPotentialEnergy; };

float IntegratorNVU::GetLagrangeMultiplier(){ 
  cudaMemcpy( h_constraintForce, d_constraintForce, sizeof(float), cudaMemcpyDeviceToHost );
  return h_constraintForce[0] / displacementLength;
} 

float IntegratorNVU::GetEuclideanLength(){ 
  S->particleData.CopyPosFromDevice();
  double distanceSq = 0;

  for(unsigned i=0; i < S->num_part; i++){
    double diffX = h_euclideanConfiguration[i].x - S->particleData.h_r[i].x;
    double diffY = h_euclideanConfiguration[i].y - S->particleData.h_r[i].y;
    double diffZ = h_euclideanConfiguration[i].z - S->particleData.h_r[i].z;

    distanceSq += diffX * diffX + diffY * diffY + diffZ * diffZ;
  }

  return sqrt(distanceSq);
} 

double IntegratorNVU::GetKineticEnergy(){ return -1; } // Not implemented.


