/*
    Copyright (C) 2010  Thomas Schrøder

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    LICENSE.txt file for license details.
*/

#include "md_NeighborList.h"

#include "rumd_technical.h"

#include "md_kernel.h"

#include "rumd_utils.h"
#include <cassert>

//////////////////////////////////////////////////////////////////
///  CONSTRUCTION/DESTRUCTION, MEMORY ALLOCATION/DEALLOCATION ////
//////////////////////////////////////////////////////////////////


NeighborList::NeighborList(int set_con){
  
  con = set_con;

  // these need to be zero before Allocate functions are first called
  numBlocks = 0;
  numBlocksParticles = 0;
  
  // initialize array pointers to null
  NbList = 0;

  // allocate memory not dependent on the number of particles

  // cut-offs
  CUDA_SAFE_CALL( cudaMalloc( (void**) &d_shell2, sizeof(float)*MaxNumTypes*MaxNumTypes ) );
}

NeighborList::~NeighborList() {
  Free();
  CUDA_SAFE_CALL(cudaFree(d_shell2));
}

void NeighborList::Initialize(ParticleData* pData, dim3 set_grid, dim3 set_threads, size_t set_shared_size, int nb, int nvp) {

  particleData = pData;

  numParticles = particleData->numParticles;
  
  grid.x = set_grid.x;
  grid.y = set_grid.y;
  grid.z = set_grid.z;

  threads.x = set_threads.x;
  threads.y = set_threads.y;
  threads.z = set_threads.z;

  shared_size = set_shared_size;

  printf("shared_size %zu\n",shared_size);
  printf("numParticles %d\n",numParticles);
  int nb_nvp = nb*nvp;

  if( numBlocks == nb && numBlocksParticles == nb_nvp)
    return;
  Free();

  CUDA_SAFE_CALL( cudaMalloc( (void**) &NbList, sizeof(NBTYPEv) * nb_nvp ) );
  CUDA_SAFE_CALL( cudaMalloc( (void**) &cell_list, sizeof(int)  * nb*nb ) );  
  
  numBlocksParticles = nb_nvp;
  
  // for checking whether to update
  CUDA_SAFE_CALL( cudaMalloc( (void**) &d_ctrl, sizeof(int)*nb ) );
  numBlocks = nb;
}


void NeighborList::Free() {
  if(numBlocksParticles != 0)
    {
      CUDA_SAFE_CALL(cudaFree(NbList));
    }
  
  if(numBlocks != 0) {
    CUDA_SAFE_CALL(cudaFree(d_ctrl));
    CUDA_SAFE_CALL(cudaFree(cell_list));
  }
}

void NeighborList::SetCutoffs(Potential* potential, float skin) {

  for (unsigned int i=0; i<MaxNumTypes; i++) {
    for (unsigned int j=0; j<MaxNumTypes; j++) { 
       unsigned int k = i * MaxNumTypes + j;
       double Rcut = sqrt( (double)potential->h_params[i][j][0] );
       potential->h_shell2[k] = ( Rcut + skin ) * ( Rcut + skin );
    }
  }  

  CUDA_SAFE_CALL( cudaMemcpy( d_shell2, potential->h_shell2, sizeof(float)*MaxNumTypes*MaxNumTypes, cudaMemcpyHostToDevice) );
}

void NeighborList::UpdateNBlist(float* d_L, float* d_invL){ 
  float4* d_r = particleData->d_r;
  float4* d_im = particleData->d_im;

  md_update_nblist<<< grid, threads, shared_size >>>( numParticles, d_r, d_im, d_shell2, d_L, d_invL, NbList, cell_list, d_ctrl );
}


void NeighborList::UpdateNBlistInit(float* d_L, float* d_invL) {  
  float4* d_r = particleData->d_r;
  float4* d_im = particleData->d_im;
  
  md_update_nblist0<<< grid, threads, shared_size >>>( numParticles, d_r, d_im, d_shell2, d_L, d_invL, NbList, cell_list, d_ctrl );
  //PrintCellList();
}

void NeighborList::PrintCellList() {
  unsigned int nb = numBlocks;
  int* h_cell_list; 
  int EmptyCells = 0;

  h_cell_list = (int*) malloc( sizeof(int) * nb * nb );
  CUDA_SAFE_CALL( cudaMemcpy( h_cell_list, cell_list, sizeof(int)*nb*nb, cudaMemcpyDeviceToHost) );

  //printf("cell_list:\n"); 
  for (unsigned int i=0; i<nb; i++) {
    //printf("%d: ", i);
    for (unsigned j=0; j<nb; j++) {
      //printf("%d ", h_cell_list[i*nb + j]);
      if (h_cell_list[i*nb + j]<0) EmptyCells++;
    }
    //printf("\n");
  }
  //printf("\n");

  //printf("Empty cells: %d, Density of empty cells %f\n", EmptyCells, float(EmptyCells)/float(nb*nb));
  printf(" %f\n", float(EmptyCells)/float(nb*nb));

  free(h_cell_list);
}
