#ifndef MD_IO_HELPERS_H
#define MD_IO_HELPERS_H

/*
    Copyright (C) 2010  Thomas Schrøder

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    LICENSE.txt file for license details.
*/

#include <cstdio>
#include <cassert>
#include <cstdlib>

#include <string>
#include <map>

#include "rumd_base.h"


// this struc handles the "logarithmic saving"
typedef struct LogLinManager{
  unsigned long int block;
  unsigned long int nextTimeStep;
  unsigned long int base;
  unsigned long int index;
  unsigned long int maxIndex;
  unsigned long int maxInterval;
} LogLin;


// this class handles two kinds of metaData for configuration files
// (1) specific data in the comment line, such as the box-length (this is
// specfically needed for reading, not so much for writing since it is available
// to the sample class anyway)
// (2) which per-particle data is in the file (specifies as the "columns" entry 
// in the comment line)

class ConfMetaData
{
public:
  ConfMetaData() :
    precision(6),
    ioformat(1), // need to link this with global ioformat, maybe a define in the basic header
    numTypes(1), // refers to a file
    timeStepIndex(0),
    numMolecules(0)
  {
    // might want to remove this and place the burden of initialization
    // on the calling function
    for(unsigned int cd = 0; cd < DIM; ++cd)
      L[cd] = 0.;
    Ps = 0.;
    Pv = 0.;
    dt = 0.;
    for(unsigned int tdx = 0;tdx < MaxNumTypes; ++tdx)
      massOfType[tdx] = 1.;
    has_images = false;
    has_velocities = false;
    has_forces = false;
    has_pot_energies = false;
    has_virials = false;		
    has_moleculeIndices = false;
    logLin.block = 0;
    logLin.nextTimeStep = 0;
    logLin.base = 1;
    logLin.index = 0;
    logLin.maxIndex = 0;
    logLin.maxInterval = 0;
    blockSize = 0;
    has_logLin = false;
  }
  unsigned int precision;
  // might want to add a separate precision for energies/virials
  unsigned int ioformat;
  unsigned int numTypes;
  unsigned int timeStepIndex;
  unsigned int numMolecules;
  float L[DIM];
  float Ps;
  float Pv;
  float dt;
  float massOfType[MaxNumTypes];
  bool has_images;
  bool has_velocities;
  bool has_forces;
  bool has_pot_energies;
  bool has_virials;
  bool has_moleculeIndices;
  bool has_logLin;
  unsigned int blockSize;
  LogLin logLin;
};

class OutputFormat{
 public:
  OutputFormat():
    precision(6),
    writeInterval(100),
    usingAtomicVirPress(true)
  {
    options["potentialEnergy"] = true;
    fileStr["potentialEnergy"] = "pe";
    options["kineticEnergy"] = true;
    fileStr["kineticEnergy"] = "ke";
    options["virial"] = true;
    fileStr["virial"] = "W";
    options["totalEnergy"] = true;
    fileStr["totalEnergy"] = "Etot";
    options["temperature"] = true;
    fileStr["temperature"] = "T";
    options["pressure"] = true;
    fileStr["pressure"] = "p";	
    options["volume"] = false;
    fileStr["volume"] = "V";
    options["thermostat_Ps"] = true;
    fileStr["thermostat_Ps"] = "Ps";
    options["barostat_Pv"] = false;
    fileStr["barostat_Pv"] = "Pv";
    options["enthalpy"] = false;
    fileStr["enthalpy"] = "H";
    options["potentialVirial"] = false;
    fileStr["potentialVirial"] = "pot_vir";
    options["stress_xx"] = false;
    fileStr["stress_xx"] = "sxx";
    options["stress_yy"] = false;
    fileStr["stress_yy"] = "syy";
    options["stress_zz"] = false;
    fileStr["stress_zz"] = "szz";
    options["stress_yz"] = false;
    fileStr["stress_yz"] = "syz";
    options["stress_xz"] = false;
    fileStr["stress_xz"] = "sxz";
    options["stress_xy"] = false;
    fileStr["stress_xy"] = "sxy";
    options["v_com_x"] = false;
    fileStr["v_com_x"] = "v_com_x";
    options["v_com_y"] = false;
    fileStr["v_com_y"] = "v_com_y";
    options["v_com_z"] = false;
    fileStr["v_com_z"] = "v_com_z";
    options["configTemperature"] = false;
    fileStr["configTemperature"] = "configT";
    options["laplacian"] = false;
    fileStr["laplacian"] = "laplacian";
    options["lagrangeMultiplier"] = false;
    fileStr["lagrangeMultiplier"] = "langMulti";
    options["simulationDisplacementLength"] = false;
    fileStr["simulationDisplacementLength"] = "dispLength";
    options["displacementLengthCorrection"] = false;
    fileStr["displacementLengthCorrection"] = "dispCorrection";
    options["constraintForceCorrection"] = false;
    fileStr["constraintForceCorrection"] = "forceCorrection";
    options["instantaneousTimeStepSq"] = false;
    fileStr["instantaneousTimeStepSq"] = "dt^2";
    options["euclideanLength"] = false;
    fileStr["euclideanLength"] = "euclideanL";
  }
  void Set(std::string key, bool val) {
    ValidateOption(key);
    options[key] = val;
  }
  bool Get(std::string key) {return options[key];}
  std::map<std::string, bool>::const_iterator Start() const {return options.begin();}
  std::map<std::string, bool>::const_iterator End() const {return options.end();}
  std::string FileString(std::string key) {return fileStr[key];}

  int precision;
  int writeInterval;
  bool usingAtomicVirPress;
  // Optional items (beyond pe, ke and virial)
protected:
  std::map<std::string, bool> options;
  std::map<std::string, std::string> fileStr;


  void ValidateOption(std::string key) 
  {
    if(options.find(key) == options.end()) {
      fprintf(stderr, "Unrecognized key in OutputFormat: %s\n", key.c_str());
      exit(1);
    }
    bool ok = true;
    // the following avoid duplicate data-columns
    if(usingAtomicVirPress) {
      if(options["atomicVirial"]) ok = false;
      if(options["pressure"] && options["atomicPressure"]) ok = false;
    }
    else {
      if(options["molecularVirial"]) ok = false;
      if(options["pressure"] && options["molecularPressure"]) ok = false;
    }

    // should be something which checks the condition of having
    // no constraints and printing out the potential virial separately
    // since it would be the same as the default virial

    if(!ok) {
      fprintf (stderr,"Invalid options for OutputFormat %s\n",key.c_str());
      exit(1);
    }
  }
};

// the function which parses the comment line for configuration files
void ConfMetaReader(char *commentLine, ConfMetaData *metaData, int verbose);

// the function which parses the comment line for configuration files
void EnergiesMetaReader(char *commentLine, OutputFormat *outputFormat, int verbose);

#endif // MD_IO_HELPERS_H
