#ifndef MD_KERNEL_FORCES_H
#define MD_KERNEL_FORCES_H

/*
    Copyright (C) 2010  Thomas Schrøder

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    LICENSE.txt file for license details.
*/

#include "rumd_technical.h"

#ifndef D_PARAMS
extern float d_params[MaxNumTypes][MaxNumTypes][NumParam];
#endif

template<int STR, int CONFT, class P> __device__ __host__ float fij( P* Pot, float4 my_r, float4 rj, float4* my_f, float4* my_w, float4* my_sts, 
								     float4* my_misc, float* param, float L, float invL ){

  float dr_x = my_r.x - rj.x;
  float dr_y = my_r.y - rj.y;
  float dr_z = my_r.z - rj.z;
  
  dr_x -= L*rintf(dr_x*invL);
  dr_y -= L*rintf(dr_y*invL);
  dr_z -= L*rintf(dr_z*invL);
  
  float dist2 = dr_x * dr_x + dr_y * dr_y + dr_z * dr_z;           
  
  // Inside cut-off for interaction? (param[0]=Rcut^2)
  if ( dist2 <= param[0] & dist2 >= 0.000001f ){
    float s;
    
    if(CONFT){
      s = Pot->ComputeInteraction(dist2, param, my_f, my_w, my_misc);
    }
    else{
      s = Pot->ComputeInteraction(dist2, param, my_f, my_w);
    }
    
    (*my_f).x += dr_x * s;
    (*my_f).y += dr_y * s;
    (*my_f).z += dr_z * s;
    
    if(STR) {
      // stress - diagonal components
      (*my_sts).x -= dr_x * dr_x * s;  // xx
      (*my_sts).y -= dr_y * dr_y * s;  // yy
      (*my_sts).z -= dr_z * dr_z * s;  // zz
      // stress - off-diagonal components
      (*my_sts).w -=  dr_y * dr_z * s; // yz
      (*my_w).y -=  dr_x * dr_z * s;   // xz
      (*my_w).z -=  dr_x * dr_y * s;   // xy      
    }
  } // END  dist2 <= param[0]...
  
  return dist2;                                                 
}

template<int STR, int CONFT, class P> __device__ void calc_f( P *Pot, float4 my_r, float4 *r, float4 *f, float4 *w, float4 *sts, float4* misc, 
							      unsigned int my_type, float L, float invL, NBTYPEv *NbList ){
  
  extern __shared__ float4 array[];
  
  float4* s_r = (float4*) array;
  
  float4 my_f = {0.0f, 0.0f, 0.0f, 0.0f};
  float4 my_w = {0.0f, 0.0f, 0.0f, 0.0f};
  float4 my_sts = {0.0f, 0.0f, 0.0f, 0.0f};
  float4 my_misc = {0.0f, 0.0f, 0.0f, 0.0f};
  
  const unsigned int offset = MyT*PPerBlock;           
  const unsigned int CellOffset = MyB*NumBlocks*PPerBlock; 
  unsigned int Pread, j, cum_j;
  NBTYPE NbListCell;
  
  // Loop over all cells
  for( unsigned int cell0 = 0; cell0 < NumBlocks; cell0+=TPerPart ){
    unsigned int cell = cell0 + MyT;
    
    Pread = __umul24(cell,PPerBlock) + MyP;
    
    // Read coordinates from global to shared mem. NOTE: must be same method as in nb-list!!!
    if ( cell < NumBlocks ){
      s_r[MyP+offset] = r[Pread]; 
    }
    
    __syncthreads();  
    
    if (cell<NumBlocks){
      // Read the NbList (coded as a binary mask) from global memory to register 
      NbListCell = NbList[CellOffset + Pread].x;
      
      // Run through NbList (i.e. the '1'-bits in NbListCell)   
      cum_j = 0; cum_j--; 
      
      while (NbListCell>0) {
	cum_j += (j = __ffs(NbListCell));         // find next '1'-bit (from the right)
	NbListCell = (NbListCell >> 1) >> j-1;    // remove the '1'-bit we just found
	
	int index = offset + cum_j;
	
	fij<STR, CONFT>( Pot, my_r, s_r[index], &my_f, &my_w, &my_sts, &my_misc, d_params[my_type][__float_as_int(s_r[index].w)], L, invL );
      }  
    } // END if (cell<NumBlocks)
    __syncthreads();  
    
  } // END for( unsigned int cell0 = 0; ... 
  
  // Sum the force for all local threads. 'Borrowing s_r'
  s_r[MyP+offset] = my_f;
  
  __syncthreads();  
  
  if( MyT == 0 ){
    for( int i=1; i < TPerPart; i++ ){
      int j = MyP + __umul24(i,PPerBlock); 
      my_f.x += s_r[j].x;
      my_f.y += s_r[j].y;
      my_f.z += s_r[j].z;
      my_f.w += s_r[j].w;
    }
    
    f[MyGP] =  my_f;
  }
  
  __syncthreads();  
  // Sum the atomic/molecular virial for all local threads.
  s_r[MyP+offset] = my_w;
  __syncthreads();  
  
  if( MyT == 0 ){
    for ( int i=1; i < TPerPart; i++ ){
      int j = MyP + __umul24(i,PPerBlock); 
      my_w.w += s_r[j].w;

      if(STR){
	my_w.y += s_r[j].y;
	my_w.z += s_r[j].z;
      }
    }
    w[MyGP] =  my_w;
  }
  
  if(STR){
    
    __syncthreads();  
    // Sum the shear stress for all local threads.
    s_r[MyP+offset] = my_sts;
    __syncthreads();  
    
    if( MyT == 0 ){
      for ( int i=1; i < TPerPart; i++ ){
	int j = MyP + __umul24(i,PPerBlock); 
	my_sts.x += s_r[j].x;
	my_sts.y += s_r[j].y;
	my_sts.z += s_r[j].z;
	my_sts.w += s_r[j].w;
      }
      
      sts[MyGP] =  my_sts;
    }
  } // END if(STR)
  
  if(CONFT){
    
    __syncthreads();  
    s_r[MyP+offset] = my_misc;
    __syncthreads();  
    
    if( MyT == 0 ){
      for ( int i=1; i < TPerPart; i++ ){
	int j = MyP + __umul24(i,PPerBlock); 
	my_misc.x += s_r[j].x;
      }
      
      misc[MyGP] =  my_misc;
    }
  }
} // END calc_f


template<int STR, int CONFT, class P> __global__ void md_calc_f( P *Pot, float4* r, float4* f, float4* w, float4* sts, float4* misc, float* L, float* invL, NBTYPEv* NbList ){
  // All shared array start at the same memory location.
  extern __shared__ float4 array[];
  
  float4* s_r = (float4*) array;
  
  // One thread per particle reads the position
  if ( MyT == 0 ){
    s_r[MyP] = r[MyGP];
  }
  
  __syncthreads();
  
  float4 my_r = s_r[MyP]; 
  
  __syncthreads(); 
  
  // Calculate and save force on this particle
  calc_f<STR, CONFT>( Pot, my_r, r, f, w, sts, misc, __float_as_int(my_r.w), L[0], invL[0], NbList );
}

#endif // MD_KERNEL_FORCES_H
