/*
    Copyright (C) 2010  Thomas Schrøder

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    LICENSE.txt file for license details.
*/

#include "rumd_base.h"
#include "rumd_technical.h"

// Nearest periodic distance between particle i and j.
__device__ float calc_dist2( float4 my_r, float4 rj, float L, float invL ){
  
  float dr_x = my_r.x - rj.x;
  float dr_y = my_r.y - rj.y;
  float dr_z = my_r.z - rj.z;
  
  dr_x -= L*rintf(dr_x*invL);
  dr_y -= L*rintf(dr_y*invL);
  dr_z -= L*rintf(dr_z*invL);
  
  return (dr_x*dr_x + dr_y*dr_y + dr_z*dr_z);               
}

__device__ void update_nblist( unsigned int num_part, float4 my_r, float4 *r, float* params, float L, float invL, NBTYPEv *NbList, int *cell_list ){
  
  extern __shared__ float4 s_r[];
  unsigned int* s_NbList = (unsigned int*) s_r;

  const unsigned int offset = MyT*PPerBlock;
  const unsigned int CellOffset = MyB*NumBlocks*PPerBlock; 
  unsigned int Pread;
  NBTYPE NbListCell;
  unsigned int CellListCount = 0;
  float dist2;
  
  // Loop over all cells
  for (unsigned int cell0 = 0; cell0 < NumBlocks; cell0+=TPerPart) {
    
    unsigned int cell = cell0 + MyT;
    
    NbListCell = 0;
    Pread = __umul24(cell,PPerBlock) + MyP;
    
    if (cell<NumBlocks) {
      if (Pread<num_part) {
	s_r[MyP+offset] = r[Pread]; 
      } else {
	s_r[MyP+offset].w = -1.0f; // Mark particle as being 'dummy'
      }
    }    
    
    // Make sure all shared data loaded before using
    __syncthreads();  
    
    // Check all interactions in this cell
 if (cell<NumBlocks){
      if( MyGP < num_part ){ // If virtual just write 0.
	for (unsigned int i = 0; i < PPerBlock; i++){
	  float4 sr = s_r[i+offset];
	  if (sr.w>-0.5f) {
	    dist2 = calc_dist2(my_r, sr, L, invL);
	    
	    if ( dist2 <= params[__float_as_int(sr.w)] && dist2 >= 0.00001f ){
	      NbListCell += (1<<i); 
	    }
	  }
	}
      }
      // Write the nb-list for this cell to global memory
      NbList[CellOffset + Pread].x = NbListCell;
    }
    
    // Check if there are any interactions with this cell (borrowing shared memory)
    __syncthreads();
    if (MyP==0) s_NbList[MyT] = 0;
    __syncthreads();
    atomicOr(&(s_NbList[MyT]), NbListCell);	// Expensive?
    __syncthreads();

    if (MyP==0 & MyT==0) {
      for (unsigned j=0; j<TPerPart; j++) {
      	if (s_NbList[j]>0) {
	  cell_list[MyB*NumBlocks + CellListCount] = cell0 + j;
	  CellListCount++;	  
	}  
      }  	 
    } 
    __syncthreads();     

  }

  // Mark end of cell_list array by -1
  if (MyP==0 & MyT==0) {	 
    for (unsigned j=MyB*NumBlocks + CellListCount; j<(MyB + 1)*NumBlocks; j++) {
      cell_list[j] = -1;
    }
  }  

}

__global__ void md_update_nblist( unsigned int num_part, float4* r, float4* im, float* params, float* L, float* invL, NBTYPEv *NbList, int *cell_list, int *check ){
  if(check[0]>0){
    
    // Read LJ-parameters into shared memory
    __shared__ float s_params[MaxNumTypes*MaxNumTypes]; 
    
    if (MyP<MaxNumTypes*MaxNumTypes & MyT==0){ 
      s_params[MyP] = params[MyP];
    }
    
    extern __shared__ float4 s_r[];
    float4 my_r;
    if (MyT==0) { 
      my_r = r[MyGP];
      s_r[MyP] = my_r;
    }
    __syncthreads();
    
    my_r = s_r[MyP];
    __syncthreads();
    
    // Update nb-list for this particle
    update_nblist( num_part, my_r, r, &s_params[MaxNumTypes*__float_as_int(my_r.w)], L[0], invL[0], NbList, cell_list );
    
    // Reset nblist checks
    im[MyGP].w = 0.0f;
    //if (MyGP==0 & MyT==0) check[0]=0;
    if (MyP==0 & MyT==0) atomicSub(check,1);
  }
}

__global__  void md_update_nblist0( unsigned int num_part, float4* r, float4* im, float* params, float* L, float* invL, NBTYPEv *NbList, int * cell_list, int* check ){
  
  // Read LJ-parameters into shared memory
  __shared__ float s_params[MaxNumTypes*MaxNumTypes]; 
  
  if( MyP < MaxNumTypes*MaxNumTypes & MyT==0 ){ 
    s_params[MyP] = params[MyP];
  }
  
  // Reset nblist checks
  im[MyGP].w = 0.0f;
  
  if (MyGP==0 & MyT==0) check[0]=0;
  
  extern __shared__ float4 s_r[];
  float4 my_r;
  
  if( MyT == 0 ){ 
    my_r = r[MyGP];
    s_r[MyP] = my_r;
  }
  
  __syncthreads();
  
  my_r = s_r[MyP]; 
  
  __syncthreads(); 
  
  // Update nb-list for this particle
  update_nblist(num_part, my_r, r, &s_params[MaxNumTypes*__float_as_int(my_r.w)], L[0], invL[0], NbList, cell_list );
}
