/*
    Copyright (C) 2010  Thomas Schrøder

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    LICENSE.txt file for license details.
*/

#include "md_IntegratorNoseHoover.h"
#include "md_IntegratorNVU.h"

void Sample::ReadRestartConf( int restartBlock ) {
  // following works only on CUDA 2.2 or higher
  std::ostringstream restartFilePath;
  restartFilePath << trajectoryDir << "/restart" << std::setfill('0') 
		  << std::setw(4) << std::setprecision(4)  
		  << restartBlock << ".xyz.gz";
  
  ReadConf(restartFilePath.str());
}

void Sample::ReadConf( const std::string& filename ) {
	
  char strWorkspace[256];
  char *ok; // for checking success of fgets
  unsigned int maxLen = sizeof(strWorkspace)/sizeof(char);
  unsigned int tdx;
  int nItems;
  unsigned int set_num_part = 0;
  float4 zero = { 0.f, 0.f, 0.f, 0.f };
  float4 L = zero;

  gzFile gp;
  gp = gzopen(filename.c_str(),"r");
  if (!gp){
    fprintf (stderr, "ReadConf: Error opening file %s: %s\n",filename.c_str(), strerror(errno));
    exit(1);
  }

  // read the first line
  ok = gzgets(gp, strWorkspace, maxLen);
  if(!ok) {
    fprintf(stderr, "ReadConf: Error reading first line of file %s\n", filename.c_str());
    exit(1);
  }
  // read number of particles, do associated memory alloc
  nItems = sscanf(strWorkspace, "%u", &set_num_part);
  if(nItems != 1) {
    fprintf (stderr, "ReadConf: Error reading number of particles from file %s\n",filename.c_str());
    exit(1);
  }
  assert(set_num_part > 0);   
  SetNumberOfParticles(set_num_part);

  float4* h_r = particleData.h_r;
  float4* h_v = particleData.h_v;
  float4* h_vb = particleData.h_vb;
  float4* h_im = particleData.h_im;
  float4* h_f = particleData.h_f;
  float4* h_w = particleData.h_w; 
  unsigned int* h_Type = particleData.h_Type;

  // read the comment line
  ok = gzgets(gp, strWorkspace, maxLen);
  if(!ok || strlen(strWorkspace) == maxLen - 1) {
    fprintf(stderr, "ReadConf: Error reading comment line in file %s\n",filename.c_str());
    exit(1);
  }
  else {
    // replace final newline with a zero
    int lenCmtLine = strlen(strWorkspace);
    assert(strcmp(&(strWorkspace[lenCmtLine-1]),"\n") == 0);
    strWorkspace[lenCmtLine-1] = 0;
  }
  
  ConfMetaData metaData;
  // initialize with known parameters (recall not all parameters need to be
  // given on the comment line)
  metaData.Pv = h_Pv[0];
  for(tdx = 0; tdx < MaxNumTypes; ++tdx)
    massOfType[tdx] = 1.0; // default value
  
  // read data from comment line into metaData
  ConfMetaReader(strWorkspace, &metaData, 1);
  this->numTypes = metaData.numTypes;
  
  // assert that there were L values in the file
  L.x = metaData.L[0];
  L.y = metaData.L[1];
  L.z = metaData.L[2];
  assert(L.x > 0. && L.y > 0 && L.z > 0.); //would be zero if not read
  // note only L.x is used for now (argument to InitializeBoxVariables)
  
  // masses, Ps and Pv are allowed to be determined by the comment line
  for(tdx = 0; tdx < numTypes; ++tdx)
    massOfType[tdx] = metaData.massOfType[tdx];

  IntegratorNoseHoover* testItg = dynamic_cast<IntegratorNoseHoover*>(itg);

  if(testItg){
    testItg->SetThermostatState(metaData.Ps);    
  }

  h_Pv[0] = metaData.Pv;
  int numMolecules = metaData.numMolecules; // local.
  
  // this is required for the memcpy trick with Types
  assert(sizeof(int) == sizeof(float));
  
  // read the per-particle
  for ( unsigned int i = 0; i < num_part; i++ ){
    ok = gzgets(gp, strWorkspace, maxLen);
    if(!ok || strlen(strWorkspace) == maxLen - 1) {
      fprintf(stderr, "Error reading line %u from configuration file\n", i);
      exit(1);
    }
    int offset = 0;
    int partial_offset = 0;
    
    // type and positions
    sscanf(strWorkspace+offset, "%d %f %f %f%n", &h_Type[i], &h_r[i].x, &h_r[i].y, &h_r[i].z, &partial_offset);
    offset += partial_offset;
    // set mass
    if (h_Type[i] >= numTypes) {
      printf("%d %f %f %f %d\n", h_Type[i], h_r[i].x, h_r[i].y, h_r[i].z,numTypes );
      assert(h_Type[i] < numTypes);	    
    }
    // For sending to the device, store type as a float instead of uint

    memcpy((void*)&(h_r[i].w), (void*)&(h_Type[i]),sizeof(int));

    h_v[i].w = 1.0 / massOfType[h_Type[i]];
    h_im[i] = zero; // initialize images
    if(metaData.has_images) {
      sscanf(strWorkspace+offset, "%f %f %f%n", &h_im[i].x, &h_im[i].y, &h_im[i].z, &partial_offset);
      offset += partial_offset;
    }
    if(metaData.has_velocities)
      {
	sscanf(strWorkspace+offset, "%f %f %f%n", &h_v[i].x, &h_v[i].y, &h_v[i].z, &partial_offset);
	offset += partial_offset;
        h_vb[i].x = h_v[i].x;
        h_vb[i].y = h_v[i].y;
        h_vb[i].z = h_v[i].z;
      }	
    if(metaData.has_forces) {
      sscanf(strWorkspace+offset, "%f %f %f%n", &h_f[i].x, &h_f[i].y, &h_f[i].z, &partial_offset);
      offset += partial_offset;
    }
    if(metaData.has_pot_energies) {
      assert(potential);
      sscanf(strWorkspace+offset, "%f%n", &h_f[i].w, &partial_offset);
      offset += partial_offset;
      h_f[i].w *= 2.0/potential->potential_factor;
    }
    if(metaData.has_virials) {
      sscanf(strWorkspace+offset, "%f%n",&h_w[i].w, &partial_offset);
      offset += partial_offset;
      h_w[i].w *= 2.;
    }
    if(metaData.has_moleculeIndices) {
      int dummy; // Trond will implement an array for this
      sscanf(strWorkspace+offset, "%d%n", &dummy, &partial_offset);
      offset += partial_offset;
      assert(0 <= dummy && dummy < numMolecules);
    }
  }
  
  gzclose(gp);
  InitializeBoxVariables(L.x);
  CopyConfToDevice();
}

void Sample::WriteConf(  const std::string &filename, ConfMetaData *metaData ) {
  UpdateNBlist();
  // meta-data such as the box-length, which has meaning independent of
  // the file-format, is taken from sample class itself. We do not check
  // the passed metaData variable for consistency
  // (ConfMetaData has fields for such data more for the purpose of reading)

  ConfMetaData *thisMetaData;
  if(metaData == 0)
    thisMetaData = &restartFormat;
  else
    thisMetaData = metaData;
  
  int precision = thisMetaData->precision;

  int molIdx = 0; //local, temporary.
  int numMolecules = 0; // likewise
  
  gzFile gp;
  std::string compressedFilename(filename);
  int len = compressedFilename.size();
  // add .gz if not already present
  if(len < 3 || compressedFilename.substr(len-3,3).compare(".gz"))
    compressedFilename.append(".gz");

  // the mode is append if this is a trajectory file and the index is nonzero  
  if(thisMetaData->has_logLin && thisMetaData->logLin.index > 0)
    gp = gzopen(compressedFilename.c_str(),"a+");
  else
    gp = gzopen(compressedFilename.c_str(),"w");

  float4 sumU, sumK, sumW, sumS, sumC;
  float stress[6];
  if(thisMetaData->has_pot_energies || thisMetaData->has_virials)
    this->CalcEnergies(&sumU, &sumK, &sumW, &sumS, &sumC, stress); // includes copy
  else
    CopyDataFromDevice(); // maybe subject to comparison of counters etc...

  float4* h_r = particleData.h_r;
  float4* h_v = particleData.h_v;
  float4* h_im = particleData.h_im;
  float4* h_f = particleData.h_f;
  float4* h_w = particleData.h_w; 
  unsigned int* h_Type = particleData.h_Type;
  
  // first line, number of particles
  gzprintf(gp, "%u\n",num_part);
  
  // second line, "comment-line", containing meta-data
  float4 L = {h_L[0], h_L[0], h_L[0], 0.};

  gzprintf(gp, "ioformat=%d dt=%.*f", ioformat, precision, itg->GetTimeStep());
  if(thisMetaData->has_logLin)
    gzprintf(gp," timeStepIndex=%u logLin=%u,%u,%u,%u,%u", thisMetaData->logLin.nextTimeStep, thisMetaData->logLin.block, 
	     thisMetaData->logLin.base,thisMetaData->logLin.index, thisMetaData->logLin.maxIndex, thisMetaData->logLin.maxInterval);

  float thermostatState = 0;
  
  IntegratorNoseHoover* testItg = dynamic_cast<IntegratorNoseHoover*>(itg);
  
  if(testItg){
    thermostatState = testItg->GetThermostatState();
  }
  
  gzprintf(gp, " boxLengths=%.*f,%.*f,%.*f numTypes=%d Nose-Hoover-Ps=%.*f Barostat-Pv=%.*f", precision, L.x, precision, L.y, precision, 
	   L.z, numTypes, precision, thermostatState, precision, h_Pv[0]); 
  
  gzprintf(gp, " mass=%.*f",precision,  massOfType[0]);
  for(unsigned int tdx = 1; tdx < numTypes; ++tdx)
    gzprintf(gp, ",%.*f", precision, massOfType[tdx]);
  
  if(thisMetaData->has_moleculeIndices)
	gzprintf(gp, " numMolecules=%d", numMolecules);

  gzprintf(gp, " columns=type,x,y,z");
  if(thisMetaData->has_images) gzprintf(gp, ",imx,imy,imz");
  if(thisMetaData->has_velocities) gzprintf(gp, ",vx,vy,vz");
  if(thisMetaData->has_forces) gzprintf(gp, ",fx,fy,fz");
  if(thisMetaData->has_pot_energies) gzprintf(gp, ",pe");
  if(thisMetaData->has_virials) gzprintf(gp, ",vir");
  if(thisMetaData->has_moleculeIndices) gzprintf(gp, ",molIdx");
  gzprintf(gp, "\n");

  for ( unsigned int i = 0; i < num_part; i++ ){
    gzprintf(gp, "%d %.*f %.*f %.*f", h_Type[i], precision, h_r[i].x, 
	     precision, h_r[i].y, precision, h_r[i].z);
    if(thisMetaData->has_images)
      gzprintf(gp, " %.0f %.0f %.0f", h_im[i].x, h_im[i].y, h_im[i].z);
    if(thisMetaData->has_velocities)
      gzprintf(gp, " %.*f %.*f %.*f", precision, h_v[i].x, 
	       precision, h_v[i].y, precision, h_v[i].z);
    if(thisMetaData->has_forces)
      gzprintf(gp, " %.*f %.*f %.*f", precision, h_f[i].x, 
	       precision, h_f[i].y, precision, h_f[i].z);
    if(thisMetaData->has_pot_energies)
      gzprintf(gp, " %.*f", precision, h_f[i].w*potential->potential_factor/2.0);
    if(thisMetaData->has_virials)
      gzprintf(gp, " %.*f", precision, h_w[i].w/2.);
    if(thisMetaData->has_moleculeIndices)
      gzprintf(gp, " %d", molIdx);
    gzprintf(gp,"\n");
    
    // nblist check
    if ( h_im[i].w > 0.5*skin ){
      printf("\nWarning: %d %f\n", i, h_im[i].w);
    }
    
    assert(h_im[i].w <= 0.50001*skin);
  }

  gzclose(gp);  
}

void Sample::SetIOBlockInfo(unsigned int blockSize, unsigned int base, unsigned int maxInterval) {
  assert(blockSize > 0);
  assert(blockSize % base == 0);
  assert(maxInterval == 0 || 
	 (blockSize % maxInterval == 0 && maxInterval % base == 0));

  trajectoryFormat.blockSize = blockSize;
  trajectoryFormat.logLin.base = base;
  trajectoryFormat.logLin.maxInterval = maxInterval;
  writers_initialized = false;
}

void Sample::InitializeFileWriters(unsigned long int step){

  int status;
  struct stat st;
  int test_dir = stat(trajectoryDir.c_str(),&st);
  bool restart = (step > 0);
  bool trajDirExists = (test_dir == 0);

  if(restart)  {
    if(!trajDirExists) {
      fprintf(stderr, "Error: Attempting to restart, but trajectory directory does not exist\n");
      exit(1);
    }
  }
  else {// ! restart
    if(trajDirExists){
      // directory exists, mv it to backup (except if step > 0, ie a restart)
      std::string backupDir = trajectoryDir + ".rumd_bak";
      
      // test if back-up exists
      test_dir = stat(backupDir.c_str(),&st);
      if(test_dir == 0) {
	// Remove back-up directory if it exists
	std::string rmCmd("rm -rf ");
	rmCmd.append(backupDir);
	system(rmCmd.c_str());
      }
      std::cout << trajectoryDir << " exists; renaming as " << backupDir << std::endl;
      status = rename(trajectoryDir.c_str(), backupDir.c_str());
      if(status == -1) {
	fprintf(stderr, "Error %d renaming directory: %s\n",errno, strerror(errno));
	exit(1);
      }
    }
    std::cout << "Creating directory " << trajectoryDir << std::endl;
        
    mode_t dir_mode = S_IRWXU | S_IRWXG | S_IROTH | S_IXOTH;
    status = mkdir(trajectoryDir.c_str(), dir_mode);
    // check if mkdir() failed
    if(status == -1){
      fprintf(stderr, "Error opening directory %s: %s\n",trajectoryDir.c_str(), strerror(errno));
      exit(1);
    }
  } // end (!restart)
  
  if(gz_energyFile) { 
    gzclose(gz_energyFile); // if calling a second time ... 
    fprintf(stderr, "WARNING: it seems Sample::InitializeFileWriters is being called a second time on this object. blocks and energies files are being restarted, overwriting will occur.\n");
    gz_energyFile = 0;
  }

  unsigned int blockSize = trajectoryFormat.blockSize;
  LogLin &logLin = trajectoryFormat.logLin;
  logLin.block = step/blockSize;
  logLin.index = 0;
  logLin.nextTimeStep = 0;
  assert(step % blockSize == 0); // don't want to deal with anything else...

  unsigned int log2_block_size = (unsigned int)(log2(float(blockSize/logLin.base)));
  unsigned int rounded_block_size = (1 << log2_block_size)*logLin.base;
  if(blockSize != rounded_block_size) {
    fprintf(stderr, "WARNING: Nominal block size %d is not a power of 2 times the base. This means effectively that the base of LogLin will be set to this value, i.e. no logarithmic saving\n", blockSize);
    logLin.base = blockSize;
    logLin.maxIndex = 1;
  }
  else {
    logLin.maxIndex =  1 + log2_block_size;
    if(logLin.maxInterval > 0)
      logLin.maxIndex = blockSize/logLin.maxInterval + (unsigned int)(log2(float(logLin.maxInterval/logLin.base)));
  }
  
  writers_initialized = true;
}

void Sample::NotifyFileWriters(unsigned long int timeStep, float dt)
{
  if(!writers_initialized)
    InitializeFileWriters(timeStep);
  
  LogLin &logLin = trajectoryFormat.logLin;
  unsigned int blockSize = trajectoryFormat.blockSize;

  // manage logarithmic blocks for configuration writing
  FILE* fp;

  if (timeStep%blockSize==logLin.nextTimeStep%blockSize) {
    if (logLin.index > 0) {
      // have an already "opened" block: append to it
      assert(ioformat == 1);

      std::ostringstream conf_filename;
      conf_filename << trajectoryDir << "/block" << std::setfill('0') << std::setw(4) << std::setprecision(4) << logLin.block << ".xyz";

      if(writeTrajectoryConfs)
	WriteConf(conf_filename.str(), &trajectoryFormat);
      
      if (logLin.index<logLin.maxIndex) {
	// Prepare for next time around
	unsigned int increment =  (1<<(logLin.index-1)) * logLin.base;
	// restrict to an increment of maxInterval if the latter is not zero
	if(logLin.maxInterval > 0 && increment > logLin.maxInterval)
	  increment = logLin.maxInterval;
	logLin.nextTimeStep += increment;
	assert(logLin.nextTimeStep <= blockSize);
	// alowed to be equal to blockSize (last configuration in the block)
	logLin.index++;
      } else {
	// Make a note that this is the last block saved (so far)
	fp=fopen((trajectoryDir+"/LastBlockSaved").c_str(),"w");
	fprintf(fp,"%lu ", logLin.block);
	fclose(fp);
	// write a restart configuration file (full precision)
	std::ostringstream restart_filename;
	restart_filename << trajectoryDir << "/restart" << std::setfill('0') << std::setw(4) << std::setprecision(4) << logLin.block << ".xyz";
	WriteConf(restart_filename.str(), &restartFormat);

	if (logLin.block%10 == 0) printf("Finished writing data block nr. %4lu\n", logLin.block);
	logLin.block++;
	if(logLin.block > maxNumIoBlocks) {
	  fprintf(stderr, "Error: maximum number of blocks reached.\n");
	  exit(1);
	}
	logLin.index = 0;
      }
    }
    if (logLin.index==0) {
      // start new block: overwrite an existing file 
      assert(ioformat == 1);
      std::ostringstream conf_filename;
      conf_filename << trajectoryDir << "/block" << std::setfill('0') << std::setw(4) << std::setprecision(4) << logLin.block << ".xyz";
      if(writeTrajectoryConfs)
	WriteConf(conf_filename.str(), &trajectoryFormat);
      
      //logLin.nextTimeStep += 1;
      logLin.nextTimeStep = logLin.base;
      logLin.index = 1;
      if(outputFormat.writeInterval != 0) {
	// also here we close the old, open a new energy file, if necessary
	if(gz_energyFile) gzclose(gz_energyFile);
	std::ostringstream energyFilename;
	energyFilename << trajectoryDir << "/energies" << std::setfill('0') << std::setw(4) << std::setprecision(4) << logLin.block << ".dat.gz";
	gz_energyFile = gzopen(energyFilename.str().c_str(),"w");
	
	newEnergyFile = true;
      } // if(outputFormat......)
    } // if logLin.index==0)
    
  } //  if(timeStep%blockSize ...)
  
  if (outputFormat.writeInterval != 0 &&
      timeStep % outputFormat.writeInterval == 0)
    WriteEnergies(dt);

}

void Sample::WriteEnergies(float dt) {
  float4 sumU, sumK, sumW, sumS, sumC;
  float stress[6];
  this->CalcEnergies(&sumU, &sumK, &sumW, &sumS, &sumC, stress);
    
  float E = sumK.w + sumU.w;
  float Tt = sumK.y;
  float V = h_V[0];
  float P_atom = ( 2.0 * sumK.w / 3.0 + sumW.w ) / V;
  float H = E + P_atom * V; 
  
  IntegratorNoseHoover* testNH = dynamic_cast<IntegratorNoseHoover*>(itg);
  IntegratorNVU* testNVU = dynamic_cast<IntegratorNVU*>(itg);
  
  float4 totalMomentum = { 0, 0, 0, 0 };

  if(testNH){
    totalMomentum = testNH->GetTotalMomentum();
  }
  
  std::map<std::string, float> Values;
  Values["potentialEnergy"] = sumU.w/num_part;
  Values["kineticEnergy"] = sumK.w/num_part;
  Values["virial"] = sumW.w/num_part;
  Values["totalEnergy"] = (sumK.w + sumU.w)/num_part;
  Values["temperature"] = sumK.y;
  Values["pressure"] =  (num_part*Tt + sumW.w)/V;
  Values["volume"] = V;
  Values["stress_xx"] = stress[0]/V;
  Values["stress_yy"] = stress[1]/V;
  Values["stress_zz"] = stress[2]/V;
  Values["stress_yz"] = stress[3]/V;
  Values["stress_xz"] = stress[4]/V;
  Values["stress_xy"] = stress[5]/V;
  
  if(testNH){
    Values["thermostat_Ps"] = testNH->GetThermostatState();
  }
  
  Values["barostat_Pv"] = this->h_Pv[0]; // Obsolete. Remove.
  Values["enthalpy"] = H/num_part;
  Values["atomicPressure"] = P_atom;
  Values["potentialVirial"] = sumW.w;
  Values["v_com_x"] = totalMomentum.x / totalMomentum.w;
  Values["v_com_y"] = totalMomentum.y / totalMomentum.w;
  Values["v_com_z"] = totalMomentum.z / totalMomentum.w;
  Values["configTemperature"] = sumC.x / sumC.y; // Expression 1
  Values["laplacian"] = sumC.y; // Expression 2 (partial)

  if(testNVU){
    Values["lagrangeMultiplier"] = testNVU->GetLagrangeMultiplier();
    Values["simulationDisplacementLength"] = testNVU->GetSimulationDisplacementLength();
    Values["displacementLengthCorrection"] = testNVU->GetDisplacementLengthCorrection();
    Values["constraintForceCorrection"] = testNVU->GetConstraintForceCorrection();
    Values["instantaneousTimeStepSq"] = testNVU->GetSimulationTimeStepSq();
    Values["euclideanLength"] = testNVU->GetEuclideanLength();
  }
    
  int precision = outputFormat.precision;
  assert (ioformat > 0);
  
  // list structure to iterate through and then print only the items which
  // are "true", for the comment line (changes the order)
  std::map<std::string, bool>::const_iterator it;
  if(newEnergyFile) {
    gzprintf(gz_energyFile, "# ioformat=%d N=%d Dt=%f columns=",ioformat,num_part,dt*outputFormat.writeInterval);
    bool haveFirst = false;
    for(it = outputFormat.Start(); it != outputFormat.End(); it++ )
      if(it->second) {
	if(haveFirst)
	  gzprintf(gz_energyFile, ",");
	
	gzprintf(gz_energyFile, "%s",outputFormat.FileString(it->first).c_str());
        haveFirst = true;
      } // if(it->second)
    gzprintf(gz_energyFile, "\n");
    newEnergyFile = false;
  }

    for(it = outputFormat.Start(); it != outputFormat.End(); it++ )
      if(it->second) {
	gzprintf(gz_energyFile, "%.*g ", precision, Values[it->first]);
      }

    gzprintf(gz_energyFile, "\n");
}
